#define _POSIX_C_SOURCE 200904L
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "simplecom.h"
#include "dslconf.h"
#include "pidof.h"
#include "server.h"

/* the executable to start/stop for the *_LOGIC requests */
#define LOGIC_CORE		"/usr/bin/cm_pc"
#define LOGIC_CORE_WRAPPER	"/usr/bin/cm_pc_wrapper.sh"
#define LOGIC_CHILD		"/usr/bin/cm_logic"

/* defined in ticfg_tool.c */
int flash_config_str(char *cstr, size_t clen);
int get_flash_checksum(void);
int get_str_checksum(const char *str, size_t len);

/** Write the given configuration to flash. */
static int flash_cfg_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	dnotify("got flash checksum: 0x%x, str checksum: 0x%x",
			get_flash_checksum(), get_str_checksum(pkt->data, ntohl(pkt->hdr.dlen)));

	if (get_flash_checksum() == get_str_checksum(pkt->data, ntohl(pkt->hdr.dlen)))
		return send_ack_msg(iface, pkt->hdr.type, "config already up to date");

	if (flash_config_str(pkt->data, ntohl(pkt->hdr.dlen)))
		return send_nack_msg(iface, pkt->hdr.type, "horror trip!!");

	return send_ack_msg(iface, pkt->hdr.type, "flash success");
}

SIMPLE_DSLCONF_CMD(FLASH_CFG, flash_cfg);

/** Manage the modem's logic - either start it initially or have it reload, depending on parameters. */
static int logic_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	static int pid = 0;

	dnotify("logic_backend got called");

	if (pkt->hdr.type == REQ_START_LOGIC) {
		/* no need to check for already running LOGIC_CORE, will abort in that case */
		/* but check anyway, as viprinet likely triggers these requests right after
		 * each other in a completely braindead manner */
		if (!(pid = pidof(LOGIC_CORE)))
			system(LOGIC_CORE_WRAPPER);
	} else if (pkt->hdr.type == REQ_RELOAD_LOGIC) {
		/* If there is a pppd running, just send it a SIGHUP (so it terminates
		 * gracefully). When killing the LOGIC_CHILD instead, pppd won't cleanly
		 * hangup the line, leading to a "session limit exceeded" error at the
		 * immediately following dial-in attempt. */
		if (!(pid = pidof(LOGIC_CORE)))
			system(LOGIC_CORE_WRAPPER);
		else if ((pid = pidof("/usr/sbin/pppd_orig"))) {
			kill(pid, SIGHUP);
			for (pid = 0; pid < 10 && pidof("/usr/sbin/pppd_orig"); pid++)
				;
		} else if ((pid = pidof(LOGIC_CHILD)))
			kill(pid, SIGTERM);
	} else if (pkt->hdr.type == REQ_STOP_LOGIC) {
		if ((pid = pidof("/usr/sbin/pppd_orig")))
			kill(pid, SIGHUP);
		for (pid = 0; pid < 10 && pidof("/usr/sbin/pppd_orig"); pid++)
			;
		if ((pid = pidof(LOGIC_CORE)))
			kill(pid, SIGTERM);
	} else
		return send_nack_msg(iface, pkt->hdr.type, "unknown packet type passed");
	return send_ack(iface, pkt->hdr.type);
}

SIMPLE_DSLCONF_CMD(START_LOGIC, logic);
SIMPLE_DSLCONF_CMD(RELOAD_LOGIC, logic);
SIMPLE_DSLCONF_CMD(STOP_LOGIC, logic);

/** Fetch and flash the firmware the given URL points to. */
static int flash_fw_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char buf[256];

	dnotify("got a fwflash request, flashing firmware from url %s", pkt->data);
	send_ack(iface, pkt->hdr.type);
	snprintf(buf, 256, "/bin/dslconf_upgrade.sh %s", pkt->data);
	system(buf);
	return 0;
}

SIMPLE_DSLCONF_CMD(FLASH_FW, flash_fw);

#define SYNC_INFO_FILE	"/proc/avalanche/avsar_modem_training"
/* content looks like:
 * | SHOWTIME
 * | 1
 * | failTrains=0
 */
static char sync_status[128];

static int sync_stat_update(void)
{
	int fd;
	char *p, inbuf[128], new_status[128];
	const char *status;

	if ((fd = open(SYNC_INFO_FILE, O_RDONLY)) == -1) {
		derror("could not open " SYNC_INFO_FILE);
		return -1;
	}

	memset(inbuf, 0, 128);
	if (read(fd, inbuf, 127) < 2) {
		close(fd);
		perror("read()");
		return -1;
	}
	close(fd);

#define FIRST_CHARS(c1, c2) ((inbuf[0] == (c1)) && (inbuf[1] == (c2)))
	if (FIRST_CHARS('S', 'H')) {	/* SHOWTIME */
		status = "SHOWTIME";
	} else if (FIRST_CHARS('I', 'N') || /* INIT */
	           FIRST_CHARS('A', 'C')) { /* ACTREQ */
		status = "INIT";
	} else if (FIRST_CHARS('I', 'D')) { /* IDLE */
		status = "IDLE";
	} else if (FIRST_CHARS('Q', 'U')) { /* QUIET1 */
		status = "QUIET";
	} else {
		derror("unexpected file content");
		return -1;
	}
#undef FIRST_CHARS

	p = &inbuf[strlen(inbuf) - 1];
	while (*p != '=')
		p--;

	snprintf(new_status, 128, "%s %s", status, p + 1);

	p = &new_status[strlen(new_status)];
	while (!isdigit(*(--p)))
		*p = '\0';

	if (strcmp(sync_status, status)) {
		memcpy(sync_status, status, 128);
		return 1;
	}
	return 0;
}

static void sync_stat_notify(struct interface iface)
{
	send_ack_msg(iface, REQ_SYNC_STAT, sync_status);
}

static int sync_stat_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	return send_ack_msg(iface, pkt->hdr.type, sync_status);
}

DSLCONF_CMD(SYNC_STAT, sync_stat);


#define RATE_INFO_FILE "/proc/avalanche/avsar_modem_stats"
/* content looks like:
 * | 
 * | AR7 DSL Modem Statistics:
 * | --------------------------------
 * | [DSL Modem Stats]
 * | 	US Connection Rate:	448	DS Connection Rate:	3456
 * | 	DS Line Attenuation:	33	DS Margin:		21
 * | 	US Line Attenuation:	19	US Margin:		26
 * | 	US Payload :		288294384	DS Payload:		1250725824
 * | 	US Superframe Cnt :	1167783	DS Superframe Cnt:	1167783
 * | 	US Transmit Power :	11	DS Transmit Power:	19
 * | 	LOS errors:		0	SEF errors:		0
 * | 	Errored Seconds:	0	Severely Err Secs:	0
 * | 	Frame mode:		3	Max Frame mode:		0
 * | 	Trained Path:		1	US Peak Cell Rate:	1056
 * | 	Trained Mode:		2	Selected Mode:		65535
 * | 	ATUC Vendor Code:	4946544E	ATUC Revision:	1
 * | 	Hybrid Selected:	1	Trellis:		1
 * | 	Showtime Count:		1	DS Max Attainable Bit Rate: 8448 kbps
 * | 	BitSwap:		1	US Max Attainable Bit Rate:	n/a
 * | 	Annex: 			AnxB	psd_mask_qualifier: 0x0100
 * | 	Power Management Status: L0	DS HLINSC: 0
 * | 	US ACTPSD: 		-345	DS ACTPSD: -364
 * | 	Total init. errors: 	0	Total init. timeouts: 0
 * | 	Showtime init. errors: 	0	Showtime init. timeouts: 0
 * | 	Last showtime init. errors: 0	Last showtime init. timeouts: 0
 * | 	ATUC ghsVid:  b5 00 49 46 54 4e 71 74
 * | 	T1413Vid: 00 00		T1413Rev: 00		VendorRev: 00
 * | 	ATUR ghsVid:  b5 00 54 53 54 43 00 00
 * | 	T1413Vid: 00 00	T1413Rev: 00	VendorRev: 00
 * | 
 * | 	[Upstream (TX) Interleave path]
 * | 	CRC: 	0	FEC: 	0	NCD: 	0
 * | 	LCD: 	0	HEC: 	0
 * | 
 * | 	[Downstream (RX) Interleave path]
 * | 	CRC: 	0	FEC: 	19	NCD: 	0
 * | 	LCD: 	0	HEC: 	0
 * | 
 * | 	[Upstream (TX) Fast path]
 * | 	CRC: 	0	FEC: 	0	NCD: 	1
 * | 	LCD: 	0	HEC: 	0
 * | 
 * | 	[Downstream (RX) Fast path]
 * | 	CRC: 	0	FEC: 	0	NCD: 	0
 * | 	LCD: 	0	HEC: 	0
 * | 
 * | [ATM Stats]
 * | 	[Upstream/TX]
 * | 	Good Cell Cnt:	6006133
 * | 	Idle Cell Cnt:	14969893
 * | 
 * | 	Tx Packets Dropped Count:	10
 * | 	Tx Bad Packets Count:	0
 * | 
 * | 	[Downstream/RX)]
 * | 	Good Cell Cnt:	26056788
 * | 	Idle Cell Cnt:	135758135
 * | 	Bad Hec Cell Cnt:	0
 * | 	Overflow Dropped Cell Cnt:	0
 * | 	Rx Packets Dropped Count:	0
 * | 	Rx Bad Packets Count:	0
 * | 
 * | 
 * | [SAR AAL5 Stats]
 * | 	Tx PDU's:	1408571
 * | 	Rx PDU's:	1736268
 * | 	Tx Total Bytes:	247614670
 * | 	Rx Total Bytes:	1207281713
 * | 	Tx Total Error Counts:	0
 * | 	Rx Total Error Counts:	0
 * | 
 * | 
 * | [OAM Stats]
 * | 	Near End F5 Loop Back Count:	0
 * | 	Near End F4 Loop Back Count:	0
 * | 	Far End F5 Loop Back Count:	0
 * | 	Far End F4 Loop Back Count:	0
 * | 	SAR OAM Ping Response Drop Count=15
 * | 
 */
static unsigned int us, ds;

static int rate_info_update(void)
{
	int fd;
	unsigned int new_us, new_ds;
	char *c, buf[256];

	if ((fd = open(RATE_INFO_FILE, O_RDONLY)) == -1) {
		derror("could not open " RATE_INFO_FILE);
		return -1;
	}


	if (read(fd, buf, 256) == -1) {
		close(fd);
		perror("read()");
		derror("failed to read from " RATE_INFO_FILE);
		return -1;
	}
	close(fd);

	/* skip past 'AR7 DSL Modem Statistics:' */
	c = buf;
	while (*c && *c != ':')
		c++;
	if (!*c) {
		derror("unexpected file content");
		return -1;
	}
	c++;
	/* skip past 'US Connection Rate:' */
	while (*c && *c != ':')
		c++;
	if (!*c) {
		derror("unexpected file content");
		return -1;
	}
	c++;
	/* read the following value */
	sscanf(c, " %u", &new_us);
	/* skip past 'DS Connection Rate:' */
	while (*c && *c != ':')
		c++;
	if (!*c) {
		derror("unexpected file content");
		return -1;
	}
	c++;
	/* read the following value */
	sscanf(c, " %u", &new_ds);

	if (new_us != us || new_ds != ds) {
		us = new_us;
		ds = new_ds;
		return 1;
	}
	return 0;
}

static void rate_info_notify(struct interface iface)
{
	char buf[256];

	snprintf(buf, 255, "%u %u", us, ds);
	send_ack_msg(iface, REQ_RATE_INFO, buf);
}

static int rate_info_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	char buf[256];

	snprintf(buf, 255, "%u %u", us, ds);
	return send_ack_msg(iface, pkt->hdr.type, buf);
}

DSLCONF_CMD(RATE_INFO, rate_info);

#define PPP_INFO_FILE	"/var/ppp_stats.txt"
static char ppp_status[128];

static int ppp_info_update(void)
{
	int fd;
	char buf[128], *p;

	if ((fd = open(PPP_INFO_FILE, O_RDONLY)) == -1) {
		derror("failed to open file " PPP_INFO_FILE);
		return -1;
	}

	memset(buf, 0, 128);
	read(fd, buf, 127);
	close(fd);

	if (!strlen(buf))
		sprintf(buf, "Unknown");

	p = buf + strlen(buf) - 1;
	while (*p < 33 || *p > 126)
		*(p--) = 0;

	if (strcmp(buf, ppp_status)) {
		memcpy(ppp_status, buf, 128);
		return 1;
	}
	return 0;
}

static void ppp_info_notify(struct interface iface)
{
	send_ack_msg(iface, REQ_PPP_INFO, ppp_status);
}

static int ppp_info_callback(struct interface iface, struct dslconf_pkt *pkt)
{
	return send_ack_msg(iface, pkt->hdr.type, ppp_status);
}

DSLCONF_CMD(PPP_INFO, ppp_info);
